/* Copyright (C) 1999 PLUS Group - University of Maryland
 *	
 * This software is provided under the terms of the GNU Lesser General Public
 * License Version 2.1, February 1999. See the header in ParkaNet.java
 * and the file License.html for details.
 */

/* Change History:
   - 05/18/00, JDH: Use a vector instead of an array to store tokens
                    in makeTokens()
   */

package ParkaClient;

import java.util.*;
import java.io.*;

/** A response object from the Parka server. 
 */
public class Response extends java.lang.Object
{
    protected  int      status;
    protected  int      size;
    protected  boolean  reqsTokeniz;
    protected  String   response;
    protected  InputStream instream;

	 //for delimiting tokens.
    protected  static final  String  DELIM = "\t";
         // The Tokenizer, used in makeTokens() in this class,
         //   applies to some commands.
         // Examples for Parka are:  GETVARIABLES, GETPREDICATES,
         //   GETKBNAMES, and GETBINDINGS
    protected  String[]  tokenList;
      // we need the following values also, with GETBINDINGS.
  // protected  int numTokens;
      // for Query responses:
  // protected  int numResponses;    //same as size, when successful query

    final static int     XCEPTION = -10;
	//when this kind of error happens, might want to use the
	//'response' field to display to ultimate user.

    /*  ************  */
 
    public Response (ParkaNet pn)
    {
	size = 0;
        status = -1;
	response = "";
	// init query related member variables:
	tokenList = null;
	// numTokens = 0;
	// numResponses = 0;    //same as size, when successful query
	instream = pn.getInStream();  //input stream assoc. with socket
    }


  /** Returns the status code of the response. */
    public int getStatus()
    {
	return  status;
    }


  /** Returns the size of the response. */
    public int getSize()
    {
	return  size;
    }


  /** Returns the string of the response. */
    public String getResponse()
    {
	return  response;
    }

  /** Compute PARKA's response:  status + " " + length + " " + response
       The format follows the parka server format for responses to
       requests, which is:
           1. status number, blank space
           2. size of the response (# chars, not including terminating '\0')
                 [server is in C code], blank space
	   3. the response

       In the case where we are getting our first response after submitting
       a query, the 1st number back is the status if it is less than zero,
       BUT, it is the number of answers if it is greater than or equal to 0.

       The 2nd number is the string length of the response, as above, and the
       3rd number is the response.

       A fuller explanation of the responses from the server is in the
       directory with the server code and description.
    */
    void computeResponse()
    {
        try
        {
            StringBuffer in = new StringBuffer();
            int tmp;
            boolean exception_occured = false;

            // Get status;  read until first space
            tmp = instream.read();
            while ((tmp != -1) && (((char)tmp) != ' '))
            {
                in.append((char)tmp);
                tmp = instream.read();
		// System.out.print((char)tmp);
            }
	    // System.out.println();

            try
            {
                status = Integer.parseInt(in.toString());
            }
            catch (NumberFormatException e)
            {
                System.out.println("Error parsing status: [" +
                                        in.toString() + "]");
		exception_occured = true;
            }
// TEST
	    // System.out.println("Status = " + status);

	    // Get size of incoming response
            in = new StringBuffer();

            tmp = instream.read();
            while ((tmp != -1) && (((char)tmp) != ' '))
            {
                in.append((char)tmp);
                tmp = instream.read();
            }

            try
            {
                size = Integer.parseInt(in.toString());
            }
            catch (NumberFormatException e)
            {
                System.out.println("Error parsing size: [" + in.toString()
                                          + "]");
		exception_occured = true;
            }
// TEST
	    // System.out.println("Size = " + size);

	    if (!exception_occured)
	    {
                // Get response
                byte[] bytearr = new byte[size];
                int readsize = 0;
                //System.out.println("Reading");
                while (readsize < size)
                {
                    readsize += instream.read (bytearr, readsize,
                                                (size-readsize));
                }
                response = new String(bytearr, 0);
            }
	    else
	    {
		response = "Communication failure with Server";
		size = response.length();
		status = XCEPTION;
	    }

            // output will be:
/* TEST
            System.out.println ("status = "
                                 + status + "; size = " + size +
                                 "; response = " + response);
*/
        }
        catch (IOException e1)
        {
            System.out.println ("IO Error in getResponse()");
	    status  = XCEPTION;
	    response = "Communication failure with Server";
	    size    =  response.length();
        }
    }

  /** Parse responses that consist of tab-delimited strings into the
    tokenList array of tokens. */
    protected void makeTokens()
    {
        int estNumTokens;
	Vector tokenVector;

	int i = 0;
	StringTokenizer  responseTokens = new StringTokenizer (response, DELIM);
	/* Counting tokens is a slow operation, so instead we store the
	   tokens in a Vector, and copy them into an array later. */
	// numTokens = responseTokens.countTokens();
        // tokenList = new String[numTokens];
	estNumTokens = response.length() / 10;
	if (estNumTokens < 100)
	  tokenVector = new Vector();
	else
	  tokenVector = new Vector(estNumTokens);

/* TEST
        System.out.println ("\nToken list (answer) is " + numTokens + " long");
*/
	while (responseTokens.hasMoreTokens())
	  {
	    // tokenList[i] = responseTokens.nextToken();
	    tokenVector.addElement(responseTokens.nextToken());
/* TEST
	    if ((i < 3) || (i + 4 >= numTokens))
	        System.out.println ("Token: " + tokenList[i]);
		  // everything, except the line above, limits tokens output
		  // to the beginning and end of the token list.
	    if ((i == 4) && (i + 4 < numTokens))
                System.out.println ("    ..............   ");
		*/
	    // i++;
	  }
	tokenList = new String[tokenVector.size()];
	tokenVector.copyInto(tokenList);
    }


  /** Returns the array of tokens populated by a previous makeTokens() call. 
   */
    public String[] getTokenList()
    {
	return   tokenList;
    }
}
