/* Copyright (C) 1999 Jeff Heflin - University of Maryland
 *	
 * This software is provided under the terms of the GNU Lesser General Public
 * License Version 2.1, February 1999. See the header in ShoeKb.java
 * and the file License.html for details.
 */

/* Change History:
   09/13/00 JDH - Added selectColumn() method
*/

package ShoeKb;

import java.util.*;

/** A set of results that is returned by a query. */
public class ResultSet {

  String[]   colVars;            // There are M bindings(col titles)
  Vector     results;             // This is an Vector (size N) of
                                            // size M value arrays
  int        numVars;


  /** Constructs a new result set for the variables in varnames. */
  public ResultSet(String[] varNames) {

    colVars = varNames;
    results = new Vector();
    if (colVars != null)
      numVars = colVars.length;
    else
      numVars = 0;
  }

  /** Constructs a new result set to hold rowCapacity rows of
    varNum variable bindings. */
  public ResultSet(String[] varNames, int rowCapacity) {

    colVars = varNames;
    results = new Vector(rowCapacity);
    if (colVars != null)
      numVars = colVars.length;
    else
      numVars = 0;
  }

  /** Constructs a new result set given an ordering of the variables
    and a matrix of variable bindings. */
  public ResultSet(String[] varOrder, String[][] bindings) {

    colVars = varOrder;
    results = new Vector(bindings.length);
    for (int i=0; i < bindings.length; i++)
      results.addElement(bindings[i]);
  }

  /** Returns the number of results in the set. */
  public int getNumResults() {

    return results.size();
  }


  /** Returns the number of variables in the result set. */
  public int getNumColumns() {

    return numVars;
  }

  /** Returns the variable name associated with the requested column. */
  public String getColVar(int col) {

    return colVars[col];
  }

  /** Returns an array of the variable names used in the query, using
    the order that the results occur in. */
  public String[] getColVars() {

    return colVars;
  }

  /** Returns column index of named variable, or -1 if error--
    eg, name not there. */
  public int getColNameIndex(String varName) {

    for (int i = 0; i < numVars; i++) {
      if (colVars[i].equals(varName))
	return i;
    }
    return -1;
  }


  /** Returns a single binding for a single variable. */
  public String getValue (int row, int col) {

    Result   res;
    String   val;
    
    res = (Result) results.elementAt(row);
    try {
      val = res.valArray[col];
    } catch (ArrayIndexOutOfBoundsException  exc) {
      System.out.println ("Failed index: " + exc.getLocalizedMessage());
      exc.printStackTrace();
      val = "";
    }

    return  val;
  }

  /** Returns one set of bindings for the query as an array. */
  public String[] getResult(int pos) {

    Result res;
    
    res = (Result) results.elementAt(pos);
    return res.valArray;
  }

  /** Returns a matrix of the result set, where each column represents
    a different variable and each row represents a different set of
    bindings. */
  public String[][] getValueMatrix()  {

    String[][] valueMatrix;

    valueMatrix = new String[results.size()][numVars];
    for (int i = 0; i < results.size(); i++) {
      valueMatrix[i] = getResult(i);
    }
    return valueMatrix;
  }


  /** Selects a particular variable (based on its order in the results)
    and returns its set of bindings as an array. */
  public String[] selectColumn(int cnum) {

    String[] col;

    col = new String[results.size()];
    for (int i = 0; i < results.size(); i++) {
      col[i] = getValue(i, cnum);
    }
    return col;
  }

  /** Removes duplicate rows from the result set. */
  /* NOTE: This method could be made more efficient if we used a
     Quicksort or some other sorting algorithm. */
  public boolean removeDuplicates()
  {
    boolean modified = false;

    // Check up to the second to last element...
    for (int i = 0; i < results.size() - 1; i++) {
      // Check to see if it appears anywhere later in the vector...
      Result res = (Result) results.elementAt(i);
      int j = i + 1;
      j = results.indexOf(res, i + 1);
      while ((j != -1) && (j < results.size())) {
	results.removeElementAt(j);     // remove duplicate
	// We could have deleted the last element...
	if (j < results.size())
	  j = results.indexOf(res, j);     // check for next possible
      }
    }
    results.trimToSize();     // shrink capacity
    return modified;
  }

  public String toString() {

    Enumeration enum;
    Result      aRow;

    StringBuffer sbuf = new StringBuffer("{");
    enum = results.elements();
    while (enum.hasMoreElements()) {
      aRow = (Result)enum.nextElement();
      sbuf.append("{" + aRow.toString() + "}");
    }
    sbuf.append("}");
    return sbuf.toString();
  }

  /** Adds a single set of variable bindings to the result set. This
   function is used to construct result sets incrementally. */
  public void addResult(String[] res) {

    if (res.length == numVars)
      results.addElement(new Result(res));
    else
      System.out.println("ERROR: result does not have same number of " +
			 "variables as the ResultSet.");
  }

}


/** Helper class that represents a single set of bindings. */
class Result {

  String[] valArray;
  
  public Result(String[] values) {

    valArray = values;
  }

  public boolean equals(Object o)
  {
    if (o instanceof Result) {
      Result res = (Result) o;
      for (int i = 0; i < valArray.length; i++) {
	if (valArray[i].equals(res.valArray[i]) == false)
	  return false;
      }
      return true;
    }
    else
      return false;
  }

  public String toString()
  {
    StringBuffer buffer = new StringBuffer();
    if (valArray.length > 0)
      buffer.append(valArray[0]);
    for (int i = 1; i < valArray.length; i++)
      buffer.append(" " + valArray[i]);
    return buffer.toString();
  }

}


